
#include "RNXVDevice.h"

#define DEBUG_LEVEL 0

#define MAX_WAIT_FOR_RESPONSE_TIME		120000

#include "Debug.h"

RNXVDevice::RNXVDevice(Stream& SerialPort) : serialPort (SerialPort) {
  /*

    Note: Supplied UART should/need not have been initialised first.

   */
}


boolean RNXVDevice::findInResponse(const char *toMatch,
                                    unsigned int timeOut = MAX_WAIT_FOR_RESPONSE_TIME) {
  /*

   */

  int byteRead;

  unsigned int timeOutTarget; // in milliseconds


  DEBUG_LOG(1, "Entered findInResponse");
  DEBUG_LOG(2, "Want to match:");
  DEBUG_LOG(2, toMatch);
  DEBUG_LOG(3, "Found:");

  for (unsigned int offset = 0; offset < strlen(toMatch); offset++) {

    // Reset after successful character read
    timeOutTarget = millis() + timeOut; // Doesn't handle timer wrapping

    while (!serialPort.available()) {
      // Wait, with optional time out.
      if (timeOut > 0) {
        if (millis() > timeOutTarget) {
          return false;
        }
      }
      delay(1); // This seems to improve reliability slightly
    }

    // We read this separately from the conditional statement so we can
    // log the character read when debugging.
    byteRead = serialPort.read();
    
    delay(1);

    DEBUG_LOG(5, "Offset:");
    DEBUG_LOG(5, offset);
    DEBUG_LOG(3, (char) byteRead);
    DEBUG_LOG(4, byteRead);

    if (byteRead != toMatch[offset]) {
      offset = 0;
      // Ignore character read if it's not a match for the start of the string
      if (byteRead != toMatch[offset]) {
        offset = -1;
      }
      continue;
    }
  }

  return true;
}



boolean RNXVDevice::responseMatched(const char *toMatch) {
  /*
   */
  boolean matchFound = true;

  DEBUG_LOG(3, "Entered responseMatched");
  unsigned int timeOutTarget;

  for (unsigned int offset = 0; offset < strlen(toMatch); offset++) {
    timeOutTarget = millis() + MAX_WAIT_FOR_RESPONSE_TIME;
    while (!serialPort.available()) {
	  if (millis() > timeOutTarget) {
	    return false;
	  }
      // Wait -- no timeout
    }
    if (serialPort.read() != toMatch[offset]) {
      matchFound = false;
      break;
    }
  }
  return matchFound;
}



#define COMMAND_MODE_ENTER_RETRY_ATTEMPTS 5

#define COMMAND_MODE_GUARD_TIME 250 // in milliseconds

boolean RNXVDevice::enterCommandMode(boolean isAfterBoot) {
  /*

   */

  DEBUG_LOG(1, "Entered enterCommandMode");

  // Note: We used to first try to exit command mode in case we were
  //       already in it. Doing this actually seems to be less
  //       reliable so instead we now just ignore the errors from
  //       sending the "$$$" in command mode.

  for (int retryCount = 0;
       retryCount < COMMAND_MODE_ENTER_RETRY_ATTEMPTS;
       retryCount++) {

    // At first I tried automatically performing the
    // wait-send-wait-send-send process twice before checking if it
    // succeeded. But I removed the automatic retransmission even
    // though it makes things  marginally less reliable because it speeds
    // up the (hopefully) more common case of it working after one
    // transmission. We also now have automatic-retries for the whole
    // process now so it's less important anyway.

    if (isAfterBoot) {
      delay(1000); // This delay is so characters aren't missed after a reboot.
    }

    delay(COMMAND_MODE_GUARD_TIME);

    serialPort.print("$$$");

    delay(COMMAND_MODE_GUARD_TIME);
    
    serialPort.println();
    serialPort.println();
    
    if (findInResponse("CMD", 1000)) {
		return true;
	}

  }
  return false;
}



void RNXVDevice::skipRemainderOfResponse() {
  /*
   */

  DEBUG_LOG(3, "Entered skipRemainderOfResponse");
	unsigned int timeOutTarget = millis() + MAX_WAIT_FOR_RESPONSE_TIME;

    while (!(serialPort.available() && (serialPort.read() == '\n'))) {
		if (millis() > timeOutTarget) {
			return;
		}
      // Skip remainder of response
    }
}


void RNXVDevice::waitForResponse(const char *toMatch) {
  /*
   */
   unsigned int timeOutTarget = millis() + MAX_WAIT_FOR_RESPONSE_TIME;
   // Note: Never exits if the correct response is never found
   while(!responseMatched(toMatch)) {
     skipRemainderOfResponse();
     if (millis() > timeOutTarget) {
		 return;
	 }
   }
}




void RNXVDevice::begin() {
  /*
   */

  DEBUG_LOG(1, "Entered RNXVDevice::begin()");

  reboot(); // Reboot to get device into known state
  setConfiguration();
}

// TODO: Create a `begin()` that allows IP etc to be supplied.



#define SOFTWARE_REBOOT_RETRY_ATTEMPTS 5

boolean RNXVDevice::reboot(boolean isAfterBoot) {
  /*

   */

  DEBUG_LOG(1, "Entered softwareReboot");

  for (int retryCount = 0;
       retryCount < SOFTWARE_REBOOT_RETRY_ATTEMPTS;
       retryCount++) {

    // TODO: Have the post-boot delay here rather than in enterCommandMode()?

    if (!enterCommandMode(isAfterBoot)) {
      return false; // If the included retries have failed we give up
    }

    serialPort.println("reboot");

    if (findInResponse("*READY*", 2000)) {
      return true;
    }
  }

  return false;
}

boolean RNXVDevice::sendCommand(const char *command,
                                 boolean isMultipartCommand = false,
                                 const char *expectedResponse = "AOK") {
  /*
   */
  DEBUG_LOG(1, "Entered sendCommand");
  DEBUG_LOG(2, "Command:");
  DEBUG_LOG(2, command);

  serialPort.print(command);

  if (!isMultipartCommand) {
    serialPort.flush();
    serialPort.println();

    // TODO: Handle other responses
    //       (e.g. autoconnect message before it's turned off,
    //        DHCP messages, and/or ERR etc)
    waitForResponse(expectedResponse);
  }

  return true;
}


void RNXVDevice::setConfiguration() {
  /*
   */
  enterCommandMode();

  // TODO: Handle configuration better
  // Turn off auto-connect
  //sendCommand("set wlan join 0");

  // TODO: Turn off server functionality until needed
  //       with "set ip protocol <something>"

  // Set server port
  //sendCommand("set ip localport ", true);
  // TODO: Handle numeric arguments correctly.
  //uart.print(serverPort);
  //sendCommand("");

  // Turn off remote connect message
  sendCommand("set comm remote 0");

  // Turn off status messages
  // sendCommand("set sys printlvl 0");

  // TODO: Change baud rate and then re-connect?

  // Turn off RX data echo
  // TODO: Should really treat as bitmask
  // sendCommand("set uart mode 0");
}


boolean RNXVDevice::join(const char *ssid) {
  /*
   */
  // TODO: Handle other authentication methods
  // TODO: Handle escaping spaces/$ in SSID
  // TODO: Allow for timeout?

  // TODO: Do we want to set the passphrase/key to empty when they're
  //       not required? (Probably not necessary as I think module
  //       ignores them when they're not required.)

  sendCommand("join ", true);
  // TODO: Actually detect failure to associate
  // TODO: Handle connecting to Adhoc device
  if (sendCommand(ssid, false, "Associated!")) {
    // TODO: Extract information from complete response?
    // TODO: Change this to still work when server mode not active
    waitForResponse("Listen on ");
    skipRemainderOfResponse();
    return true;
  }
  return false;
}


boolean RNXVDevice::join(const char *ssid, const char *passphrase,
                          boolean isWPA) {
  /*
   */
  // TODO: Handle escaping spaces/$ in passphrase and SSID

  // TODO: Do this better...
  sendCommand("set wlan ", true);

  if (isWPA) {
    sendCommand("passphrase ", true);
  } else {
    sendCommand("key ", true);
  }

  sendCommand(passphrase);

  return join(ssid);
}


#define IP_ADDRESS_BUFFER_SIZE 16 // "255.255.255.255\0"

const char * RNXVDevice::ip() {
  /*


    The return value is intended to be dropped directly
    into calls to 'print' or 'println' style methods.

   */
  static char ip[IP_ADDRESS_BUFFER_SIZE] = "";

  // TODO: Ensure we're not in a connection?

  enterCommandMode();

  // Version 2.19 of the WiFly firmware has a "get ip a" command but
  // we can't use it because we want to work with 2.18 too.
  sendCommand("get ip", false, "IP=");

  char newChar;
  byte offset = 0;

  // Copy the IP address from the response into our buffer
  while (offset < IP_ADDRESS_BUFFER_SIZE) {
    newChar = serialPort.read();

    if (newChar == ':') {
      ip[offset] = '\x00';
      break;
    } else if (newChar != -1) {
      ip[offset] = newChar;
      offset++;
    }
  }

  // This handles the case when we reach the end of the buffer
  // in the loop. (Which should never happen anyway.)
  // And hopefully this prevents us from failing completely if
  // there's a mistake above.
  ip[IP_ADDRESS_BUFFER_SIZE-1] = '\x00';

  // This should skip the remainder of the output.
  // TODO: Handle this better?
  waitForResponse("<");
  while (serialPort.read() != ' ') {
    // Skip the prompt
  }

  // For some reason the "sendCommand" approach leaves the system
  // in a state where it misses the first/next connection so for
  // now we don't check the response.
  // TODO: Fix this
  serialPort.println("exit");
  //sendCommand("exit", false, "EXIT");

  return ip;
}

boolean RNXVDevice::configure(byte option, unsigned long value) {
  /*
   */

  // TODO: Allow options to be supplied earlier?

  switch (option) {
    case RNXV_BAUD:
      // TODO: Use more of standard command sending method?
      //enterCommandMode();
      //serialPort.print("set uart instant ");
      //serialPort.println(value);
      //delay(10); // If we don't have this here when we specify the
                 // baud as a number rather than a string it seems to
                 // fail. TODO: Find out why.
      //serialPort.begin(value);
      // For some reason the following check fails if it occurs before
      // the change of SPI UART serial rate above--even though the
      // documentation says the AOK is returned at the old baud
      // rate. TODO: Find out why
      //if (!findInResponse("AOK", 100)) {
      //  return false;
      //}
      break;
    default:
      return false;
      break;
  }
  return true;
}



